package FacetList;
use strict 'vars';

use Carp qw(croak);
use FilterUtil qw(bracket_list run_command);
use FilterConfig;
use Graph::Undirected;

sub new {
  my $invokant=shift;
  my $class=ref $invokant || $invokant;
  my %arg=@_;

  croak "must define rank, elements, facets" 
    unless (defined($arg{rank}) && defined($arg{facets}) 
	    && defined($arg{elements}));
  

  return bless ( { ELEMENTS=>$arg{elements},
		   RANK=>$arg{rank},
		   FACETS=> [ fl_sort(@{$arg{facets}}) ]},$class);
}


sub rank($){
  my $self=shift;
  
  return $self->{RANK};
}
sub elements($){
  my $self=shift;
  
  return $self->{ELEMENTS};
}
sub topcom($){
  my $self=shift;
  my @lines;

  push(@lines, sprintf("%d,%d:", $self->elements,
		      $self->rank()));
  push(@lines, "{");
  push(@lines, map( bracket_list('{','}',",",
				 map($_-1, @$_)), @{$self->{FACETS}}));

  push(@lines,"}");

  my $str=join("\n",@lines)."\n";
    
  return $str;
    
}

sub ridge_graph($){
  my $self=shift;
  my $g=new Graph::Undirected;
  my $count=0;
  foreach my $f1 (@{$self->{FACETS}}){
    foreach my $f2 (@{$self->{FACETS}}){
      if (shares_ridge($f1,$f2)) {
	$g->add_weighted_edge(facet_str($f1),1,facet_str($f2));
	print STDERR "$count edges\n" if (($count++ %1000)==0);
      }
    }
  }
  return $g;
}


sub mpc($@){
  my $self=shift;
  my %options=@_;
  my @lines;

  push(@lines, sprintf("rank=%d elements=%d",
		       $self->rank(), $self->elements));
  push(@lines, "boundary={");
  push(@lines, map( bracket_list('[',']'," ",@$_), @{$self->{FACETS}}));

  push(@lines,"}");
  if (scalar(keys(%options))){
    push(@lines,"options={");
    foreach my $k (keys(%options)){
      push(@lines,"$k=$options{$k}");
    }
    push(@lines,"}");
  }
  return join("\n",@lines)."\n";
    
}


sub n_facets($){
  my $self=shift;
  return scalar(@{$self->{FACETS}})
}

sub equal($$){
  my $self=shift;
  my $other=shift;

  my (@F1,@F2);
  @F1=@{$self->{FACETS}};
  @F2=@{$other->{FACETS}};

  return 0 if (scalar(@F1) != scalar(@F2));

  for (my $i=0; $i< scalar(@F1); $i++){
    return 0 if (facet_cmp($F1[$i],$F2[$i]) !=0);
  }
  return 1;
} 

# utility routines
sub facet_str($){
  my $ref=shift;
  return bracket_list('[', ']', ',', @$ref);

}

sub facet_cmp($$){
  my ($a,$b)=@_;
  my (@F1)=@$a;
  my (@F2)=@$b;

  while (scalar(@F1) && scalar(@F2)){
    return -1 if ($F1[0] < $F2[0]);
    return 1 if ($F1[0] > $F2[0]);
    shift(@F1);
    shift(@F2);
  }
  
  return (scalar(@F1)  <=> scalar(@F2));
}

# only works for simplices;
sub shares_ridge($$){
  my ($f1,$f2)=@_;
  my $count;
  my %hash=map { $_=>1} (@{$f1});
  
  foreach my $v (@{$f2}){
    $count++ if $hash{$v};
  } 

  if ($count==scalar(@{$f2})-1) {
    return 1;
  } else {
    return 0;
  }
}
sub fl_sort(@){
  my @facets=@_;
  return sort facet_cmp @facets

}

sub from_mpc{
  my @lines=@_;
  local($_);
  my @output=run_command("$SRC_DIR/extract_boundary",@lines);

#  print STDERR @output;

  $_=shift(@output);
  my ($elements,$rank)=split(",");
  
  my @facets;
  foreach  (@output){
    push(@facets,[split(" ")]);
  }
  return new FacetList (rank=>$rank,
			elements=>$elements,
			facets=>[@facets]);
}

# note, this relies on formatting of file. fragile.
sub from_topcom(@){
  my @lines=@_;
  local($_);
  my @facets;
  my $elements=0;
  my $rank=0;
  my $params;

  chomp($params=shift(@lines));
  $params=~ s/://;

  ($elements,$rank)=split(",",$params);

  foreach (@lines){
    chomp();
    next if (m/^\{$/);
    last if (m/^\}$/);    
    s/^\{//;
    s/\},?$//;
    my @nums=map($_+1,split(","));
    $rank=scalar(@nums)+1;
    push(@facets,[@nums])
  }
  return new FacetList (rank=>$rank, elements=>$elements,
		       facets=>[@facets]);
}

sub verify_output($@){
  my $F=shift;
  my @output=@_;
  my $errcount;
  foreach my $line (@output){
    my $chi = Chirotope::from_pm($line);
    
    if (!$chi->facets()->equal($F)){
      $errcount++;
      print "ERROR, boundary mismatch for $line";
      print $chi->facets()->mpc();
    } else {
      print STDERR ".";
    }
  }
  return $errcount;
}
1;
