#include <stdlib.h>
#include <stdio.h>
#include "uni.h"
#include "combinat.h"
#include "error_message.h"
#include "binom_table.h"

static natural_t before_first(natural_t, natural_t, natural_t);

natural_t binomial(natural_t n, natural_t r){

#ifdef COMPUTE_BINOMIALS 
  natural_t above = n, below = 1, i = 0;
  if (n < 1 || r < 0 || n < r)
    return (0);
  if (r == 0)
    return (1);

  if (n < 2 * r)
    return (binomial (n, n - r));

  while (i++ < r - 1)
    {
      above *= n - i;
      below *= i + 1;
    }
  return (above / below);

#else
  if ( n> BINOM_MAX_N  || r > BINOM_MAX_R)
    err_fatal("binomial","overflow: %d choose %d",n,r);

  return binom_table[n][r];
#endif
}				


/* 
 * Improved binomial function gratefully borrowed from Prof. %%J\"urgen%%
 * Bokowski's code.
 */

natural_t old_binomial(natural_t n, natural_t r){
  natural_t above = n, below = 1, i = 0;
  if (n < 1 || r < 0 || n < r)
    return (0);
  if (r == 0)
    return (1);
  if (n < 2 * r)
    return (binomial (n, n - r));
  while (i++ < r - 1)
    {
      above *= n - i;
      below *= i + 1;
    }
  return (above / below);
}				


extern bool first_kset(natural_t n, natural_t k, natural_t *pi){
  int i;
  for (i=0; i< k; i++)
    pi[i]=i+1;
  return TRUE;
}

/* From Nijenhuis and Wilf, lexicographic order */ 
extern bool next_kset(natural_t n, natural_t k, natural_t *ar){
    
    int h,j,m1;

    if (ar[0]==n-k+1) 
      return FALSE;


    for (h=1; (h <= k)  &&       (ar[k-h] == n+1-h) ; h++);

    m1=ar[k-h];
    
    for (j=1; j<=h; j++){
	ar[k+j-h-1]=m1+j;
    }

    return TRUE;
}


/* @before_first@ computes how many permutations occur before 
 * the first one starting with @pi0@ in the lexicographic order
 * of %${{1\ldots n}\choose{k}}$%
 */
static natural_t before_first(natural_t n, natural_t k,natural_t pi0){

  err_assert(n>=0 && k>=0 &&   pi0>= 0);
  err_assert(pi0 <= n-k+1);

  return (binomial(n,k)-binomial(n-pi0+1,k));
}



/* Returns the index (starting from 0) of a kset in the lexicograpic 
 * ordering
 * @pi@ must be sorted.
 * allows zeros in @pi@ to mark deleted elements.
 */

extern natural_t kset_index( natural_t n, natural_t k, natural_t *pi){
   natural_t sum;

  err_assert(n>=0 && k<=n);

  sum=kset_index_complete(n,k,pi,0,k,0,0,0);
  
  return sum;
}

extern natural_t kset_index_complete( natural_t n, natural_t k, natural_t *pi,
				      array_index_t i, array_index_t last_i, array_index_t j,
				      natural_t last,
				      natural_t sum ){
  /* 
   * @i@ is the count of positive (i.e.) non-deleted elements encountered
   * @j@ is the array_index_t into @pi@ to start looking
   */

  if (i>=last_i)
    return sum;

  while (pi[j] <= 0){ 
    j++;			/* skip deleted elements. */
  }
  sum += before_first(n-last,k-i, pi[j]-last);
  last=pi[j];
  j++;
  
  return kset_index_complete(n,k,pi,i+1,last_i,j,last,sum);
}

/** 
 * sort a list of small integers.
 * 
 * @param n universe 1..n
 * @param k size of subset
 * @param pi1 source
 * @param pi2 destination
 */
extern void bucket_sort(int n, int k, int *pi1, int* pi2){

  int *flags=calloc(n+1,sizeof(int));	
  int i;
  int j;
  int rep;
  int bound=set_storage(k,pi1);

  for (i=0; i<  bound ; i++)
    flags[pi1[i]]++;

  j=0;
  for (i=1; i<= n; i++){
    for (rep=0; rep<flags[i]; rep++){
      pi2[j++]=i;
    }
  }

  free(flags);
}

/* same as @kset_index@, but pi need not be sorted. Cuz we sort it :-) */
extern natural_t kset_cindex( natural_t n, natural_t k, natural_t *pi){
  natural_t *pi2;
  natural_t ret;


  pi2=calloc(set_storage(k,pi),sizeof(natural_t));
  
  bucket_sort(n,k,pi,pi2);

  ret=kset_index(n,k,pi2);
  free(pi2);

  return ret;
}

/** 
 * look for element in set. Could use binary search, but sets are small
 * we shall see if it matters
 * 
 * @param k size of set
 * @param S array representing set
 * @param e element to look for
 * 
 * @return oui ou non
 */
extern bool set_contains(natural_t k, natural_t S[], natural_t e){
  int i;
  for (i=0; i<k; i++)
    if (S[i]==e) return TRUE;

  return FALSE;
}

/* From the Upper Bound Theorem for convex polytopes. */

extern natural_t Gamma(natural_t n, natural_t d){
  
  if (d%2 == 0){
    return n *binomial(n-d/2,n-d)/(n-d/2);
  } else {
    return 2*binomial(n-(d+1)/2,n-d);
  }
}

/* 
 * a @k-partition@ is k-subset, along with its complement.
 * the third array @tau[j]@ contains (-1)^@rho@ where @rho@ is
 * the number of transpositions to sort %$< \pi, \bar{\pi}_i >$%.
 *
 * this is probably lousy terminology.  
 * The table generation in basis_equation.c should probably 
 * be retro fitted to  use these routines. 
 */


extern bool first_kpartition(natural_t n, natural_t k, natural_t *pi, natural_t *pibar, natural_t *tau){
  int i;
  for (i=0; i< k; i++)
    pi[i]=i+1;
  for (i=0; i< n-k; i++){
    pibar[i]=i+1+k;
    tau[i]=1;
  }
  return TRUE;
}

extern int set_storage(natural_t k, natural_t *pi){
  int zeros=0;
  int non_zeros=0;
  while (non_zeros < k){
    if (pi[zeros+non_zeros]==0)
      zeros++;
    else
      non_zeros++;
  }
  return zeros+non_zeros;
}


extern bool next_kpartition(natural_t n, natural_t k, 
			    natural_t *pi, natural_t *pibar, natural_t *tau){
    int i,j,lambda,sign;

    if (!next_kset( n,  k,pi ))
      return FALSE;
    sign=1;
    lambda=n;

    j=n-k-1;
    for (i=k-1; i>=0 ; i--){
      while (lambda>pi[i]){
	pibar[j]=lambda--;
	tau[j]=sign;
	j--;
      }
      lambda--;
      sign *=-1;
    }

    while (lambda>0){
      pibar[j]=lambda--;
      tau[j]=sign;
      j--;
    }
    return TRUE;
  }




/** 
 * Inverse of kset_cindex: find subset corresponding to an index.
 * Note that this is not efficient.
 * 
 * @param n elements
 * @param k size of subset
 * @param index which subset 
 * @param pi where to put it.
 */
void which_kset(natural_t n,natural_t k, natural_t index, natural_t *pi){
  int i,j;

  int size=k;
  int array_index_t=0;

  for (i=0; i<k; i++){

    if (index==0 ) {
      pi[i]=array_index_t+1;
      array_index_t=pi[i];
    } else {
      for (j=n-array_index_t-size+1; j>0  ; j--){
	if (before_first(n-array_index_t,size,j) <= index)
	  break;
      }
    
      pi[i]=j+array_index_t;
      index-=before_first(n-array_index_t,size,j);
      array_index_t=pi[i];

      size--;
    }
  }


  
}



extern void random_permutation(natural_t n, natural_t *pi){
  int i;
  for (i=0; i<n; i++){
    pi[i]=i+1;
  }

  for (i=0; i<n; i++){
    int j=(int)(uni()*(n-i));
    natural_t tmp;

    tmp=pi[i+j];
    pi[i+j]=pi[i];
    pi[i]=tmp;
  }
}

extern void output_bindex(int n,int d, int b, int *pi){
  int i;
 
  which_kset(n,d,b,pi);
  output_kset(n,d,pi);
}

extern void output_kset(int n, int d, int *pi){
  int i;
  printf("[");
  for (i=0; i<d; i++){
    if (i>0) printf(",");
    printf("%d", pi[i]);
  }
  printf("]");
}

/** 
 * compute the difference of two (sorted) ksets
 * 
 * @param k cardinality
 * @param pi1 set 1
 * @param pi2 set 2
 * @param out output
 * @return number of elements in difference pi1 \ pi2
 */

int kset_diff(int k,natural_t *pi1,natural_t *pi2,natural_t *out){
  int i=0,j=0;
  int count=0;

  err_assert(k<1000);

  while (i<k && j<k){
    if (pi1[i]==pi2[j]){
      i++;
      j++;
    } else if  (pi1[i]>pi2[j]) {
      j++;
    } else if (pi1[i] <pi2[j]){
      out[count++]=pi1[i++];
    }
  }


  while (i<k)
    out[count++]=pi1[i++];

  return count;
}


/** 
 * compute the intersection of two (sorted) sets
 * 
 * @param k1 cardinality 1
 * @param pi1 set 1
 * @param k2 cardinality 2
 * @param pi2 set 2
 * @param out output (NULL to just count);
 * @return number of elements in intersection pi1 \cap pi2
 */

int set_intersect(int k1,natural_t *pi1, int k2, natural_t *pi2,natural_t *out){
  int i=0,j=0;
  int count=0;

  err_assert(k1<1000 && k2<1000);

  while (i<k1 && j<k2){
    if (pi1[i]==pi2[j]){
      if (out){
	out[count]=pi1[i++];
      } 
      count++;
      j++;
    } else if  (pi1[i]>pi2[j]) {
      j++;
    } else if (pi1[i] <pi2[j]){
      i++;
    }
  }
  return count;
}

/** Compute an "incremental" order of bases where all bases involving 1..k
    occur before those involving k+1
*/
extern void init_reorder_table(int order, 
			    natural_t n, natural_t r, array_index_t *table){

  natural_t *pi,*pibar;
  sign_t *tau;
  int i,j;

  switch(order){
  case 'i':
    pi=calloc(r,sizeof(natural_t));
    pibar=calloc(n-r,sizeof(natural_t));
    tau=calloc(n-r,sizeof(sign_t));

    j=0;
    for (i=r; i<=n ; i++){
      first_kpartition(i-1,r-1,pi,pibar,tau);
      pi[r-1]=i;
      do {
	table[j++]=kset_cindex(n,r,pi);
      } while (next_kpartition(i-1,r-1,pi,pibar,tau));
    }
    break;
  case 'L':
    for (i=0; i<binomial(n,r); i++)
      table[i]=binomial(n,r)-i-1;
    break;
  case 'l':
  default:
    for (i=0; i<binomial(n,r); i++)
      table[i]=i;
    break;
  }
}

int index_cmp(const void *ap, const void *bp){
  array_index_t a,b;
  a=*(array_index_t *)ap;
  b=*(array_index_t *)bp;
  if (a == b)
    return 0;
  if (a < b) 
    return -1;
  return 1;
}

/** 
 * compute the parity of a permuted basis with respect to sorted order.
 * 
 * @param r size of basis set
 * @param basis set of indices
 * 
 * @return $-1^k$ where $k$ is the number of inversions to sort the indices
 */
int basis_parity(int r, int *basis){
  int i,j;
  int parity=1;
  int zeros;

  if (r<=1) return 1;

  zeros=set_storage(r,basis)-r;

  j=-1;
  for (i=0; i<r-1; i++){
    j++;
    while (basis[j]<=0)
      j++;

    if (basis[j] > basis[r+zeros-1])
      parity*= -1;
  }
  return parity*basis_parity(r-1,basis);
}
