
/* $Id$ */
/**
 * @file   oms.c
 * @author David Bremner <bremner@unb.ca>
 * $Date$
 * 
 * @brief  The main program for \em oms a recursive backtracking chirotope completion program.
 * 
 * 
 */

#include <unistd.h>
#include <stdio.h>
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "basis_equation.h"
#include "boundary.h"
#include "checkpoint.h"
#include "combinat.h"
#include "common_startup.h"
#include "debug.h"
#include "globaldef.h"
#include "int_queue.h"
#include "oms.h"
#include "oms_opt.h"
#include "oms_tune.h"
#include "output.h"
#include "pluecker.h"
#include "profile.h"
#include "ridge_graph.h"
#include "sa16_array.h"
#include "signed_set.h"
#include "startup.h"
#include "statearray.h"
#include "variable_choice.h"
#ifdef USE_ZRAM
#include <z_checkpoint.h>
#endif


int oms_procsetnum;
int oms_NodeTag, oms_GlobalTag;


/** 
 * Entry point, either from ZRAM or direct.
 * 
 * @param argc number command line arguments
 * @param argv the arguments
 * @param envp the environment
 * 
 * @return 0 on success
 */
extern FILE *yyin;

int MAIN(int argc, char *argv[], char **envp){

  na_node_p local,global;
  params_p params;
  startopt_t startopt=SO_pluecker|SO_backtrack;

#if (USE_ZRAM)
  double e_time;
#endif

  prof_init();

  
  global=cs_create_global(argc,argv,startopt);
  params=na_access(global,SLOT_params);

  if (params->restart){
#ifdef USE_ZRAM    
    (void)z_bt_RestartDFExp(oms_procsetnum, 
			    (z_bt_GlobalP) global,params->sequential);
#else
    err_fatal(HERE,"Checkpointing only supported with ZRAM");
#endif
  } else {

    
    parse_input(params);

    output_init(params);
    global=cs_init_global_tables(global,startopt);
    params=na_access(global,SLOT_params);
    do{
      prof_reset();


      local=cs_alloc_local(global,startopt);
      local=oms_alloc_local(global,local);
      local=cs_init_local(global,local,startopt,oms_read_boundary_hook);
      if (local){

#ifdef USE_ZRAM
	e_time=z_Wtime();
	if (params->sequential){
	  z_bt_EnumerateDepthFirst(oms_procsetnum, (z_bt_GlobalP) global,
				   (z_NodeP)local);
	} else {
	  z_bt_EnumerateDFExp(oms_procsetnum, (z_bt_GlobalP) global,
			      (z_NodeP)local, 0);
	  
	}
#else
	err_debug(DB_STARTUP,("oms","calling oms_BranchInPlace directly"));
	(void)oms_BranchInPlace(global,local,0,0);
#endif
 
#ifdef USE_ZRAM
	  e_time=z_Wtime()-e_time;
	  fprintf(stderr,"Processors: %d Time: %f\n",z_numprocs,e_time);
	  z_DumpMemUsage();
#endif
	na_destroy(local);
      }

      prof_dump(params->profile_level);
    }    while(parse_input(params));
    output_finish(params);
    na_destroy(global); 
  }  
  return(0);

}

#if 0
/**
   Debugging routine.
*/
void dump_state(na_node_p local){
  sa16_array_p pe_state;
  
  pe_state=na_access(local,SLOT_pe_state);
  
  fprintf(stderr,"* pe_state stack %d  %d\n",pe_state->stacksize,
	  pe_state->maxtop);

}
#endif





/**
   \brief This is the routine invoked by ZRAM. It does not return 
   the state, because there is no where to return it to
*/
bool oms_BranchInPlace(na_node_p global,  na_node_p state,
			  ZRAM_ARGS_PROTO){
  na_fixup(global);
  na_fixup(state);

  (void)oms_real_BranchInPlace(global,state,ZRAM_ARGS);

  na_free_aux(state);
  return(0);
}
/** 
 * \brief Do the actual work for Branch in place, but don't fixup up state 
 * blocks from (possible) network transmission.
 * 
 * @param global Global state
 * @param state Local state
 * 
 * @return true if search should continue
 */
bool oms_real_BranchInPlace(na_node_p global,  na_node_p state,
			  ZRAM_ARGS_PROTO){
  array_index_t j;
  params_p params;
  lparams_p lparams;
  sarray_p chirotope;
  bool keep_looking=TRUE;

  NODECHECK(state);

  params=na_access(global,SLOT_params);  
  lparams=na_access(state,SLOT_lparams);
  chirotope=na_access(state,SLOT_chirotope);

  if (params->max_depth>=0 && lparams->depth>=params->max_depth){
    output_state(global,state);
    prof_event(PROF_DEPTH_LIMIT);
    return TRUE;
  }

  if (OMS_DEBUGLEVEL >= 3){
    err_warning("oms_BranchInPlace","called with state: %s",
		string_pm(chirotope,sign2char));
  }

  j=vc_choose_unset(global,state);
  if (j==0){
    /* There are no unset variables */
    if (params->tree)
      putchar('s');
    prof_event(PROF_SUCCESS);
    (void) output_state(global,state);

    if (params->max_results>=1 && params->max_results<=prof_get(PROF_SUCCESS)){
      return FALSE;
    }
  } else {
    int bindex,first_sign;
    if (params->tree) putchar('(');

    bindex=si_index(j);
    err_assert(sa_read(chirotope,bindex)==0);
    first_sign=si_sign(j);

    keep_looking=TrySetting(state,global,bindex,first_sign,ZRAM_ARGS);

    if (params->tree) printf(")(");

    if (keep_looking)
      keep_looking=TrySetting(state,global,bindex,-1*first_sign,ZRAM_ARGS);
    if (params->tree) putchar(')');
  }
  return(keep_looking);
}


/**
 * @brief Save state, try to complete with given setting and restore state.

   
This routine does the following:
  - Save current state
  - set variable
  - find consequences
  - recursively try to complete the state
  - restore the state
 * 
 * @param state local state
 * @param global global state
 * @param basis which variable to set
 * @param sign what to set it to
 * 
 * @return  true if we should keep exploring.
 */  

bool TrySetting(na_node_p state, 
		na_node_p global,
		array_index_t basis, 
		sign_t sign,
		ZRAM_ARGS_PROTO){
  
  bool consistent,keep_looking=TRUE;
  
  lparams_p lparams=na_access(state,SLOT_lparams);
  params_p params=na_access(global,SLOT_params);


  maybe_prof_event(PROF_NODE);


  err_debug(DB_BACKTRACK,("TrySetting","%d %d %d %d %d",
			  basis,sign,lparams->depth,
			  sa_zeros(na_access(state,SLOT_chirotope)),
			  lparams->subtreesize));

  SaveState(state);

  lparams->subtreesize++;

  consistent=SetAndConsequences(global,state,basis,sign);

  if (consistent){
    /* recursively try and complete the modified state */
    err_debug(DB_LOADBALANCE,(HERE,"subtree lim %d",params->subtreelimit));

    if (OPT_use_zram && (lparams->subtreesize >= params->subtreelimit)) { 

      /* compact for network transport */
      search_state new_node=na_compact(state);

      lparams=na_access(new_node,SLOT_lparams);
      lparams->subtreesize=0;

      err_debug(DB_LOADBALANCE,("TrySetting","Passing to load balancer\n"));
      (void)whatToDo((z_NodeP) new_node, ref);

      na_destroy(new_node);
      keep_looking=TRUE;

    } else {
      /* do the work directly, without going through the load balancing
       * system 
       */
      keep_looking=oms_real_BranchInPlace(global,state,whatToDo,ref);
    }
  } else {
    /*    printf("pruning: "); */
    /*    output(state); */
  }
  RestoreState(state);
  return keep_looking;
}


/** 
 * Test single variable setting for consistency.


 This function has two interrelated jobs: 
 - test if the suggested
 setting is consistent (i.e. ok with pl\"{u}cker, boundary, and diamater)
 - Find variable settings forced by this one.

 * 
 * @param state local state
 * @param global global state
 * @param basis which variable to set
 * @param sigma what sign to set it to
 * 
 * @return True if that setting is consistent (as far as we can tell)
 */
bool SetAndConsequences(			   na_node_p	global,
						   na_node_p	state,

			   array_index_t	basis, 
			   sign_t	sigma){

  sarray_p chirotope;

  be_block_p PE;		/* %%Pl\"ucker%% equations */
  sa16_array_p pe_state;
  params_p params;
  sign_t old_sign;
  signed_set_p S; /* forced variables */

  NODECHECK(state);

  chirotope=na_access(state,SLOT_chirotope);
  pe_state=na_access(state,SLOT_pe_state);
  
  PE=na_access(global,SLOT_pe);

  params=na_access(global,SLOT_params);
  
  

  /* note that this check is mostly unnecessary, except during startup */
  old_sign=sa_read(chirotope,basis);
  if (old_sign !=0 && old_sign!=sigma)
    return FALSE;

  S=ss_new(sa_size(chirotope));

  (void)ss_add(S,basis,sigma);

  while (!ss_empty(S)){
    int choice;
    int bindex,sign;

    choice=ss_deq(S);
    bindex=si_index(choice);
    sign=si_sign(choice);
    
    sa_set(chirotope,bindex,sign);


    if (!be_boundary_traverse2(bindex,
				state,global,
				S,NULL)){
      err_debug(DB_BACKTRACK,(HERE,"boundary eqns failed"));
      ss_destroy(S);
      return FALSE;  
    }
    if (!pl_traverse(bindex, 
			      PE,
			      S,chirotope,pe_state)){
      if (params->tree)
	putchar('p');
      err_debug(DB_BACKTRACK,(HERE,"pluecker eqns failed"));
      ss_destroy(S);
      return FALSE;
    }
  }
  ss_destroy(S);
  return TRUE;
}

#if (USE_ZRAM)
/**
   ZRAM initialization.
 This function is called once on every processing element. 
*/
void initproc(void)
{

  
  z_signal_Init();

  z_bt1_Init();
  z_bt2_Init();
  z_check_Init();
  oms_NodeTag = z_InstallClass(oms_Node, 0, 0, 0);
  oms_GlobalTag = z_InstallClass(oms_Global, 0, 0, 0);
  assert(oms_GlobalTag>0);
  oms_procset.bt_BranchInPlace = (z_bt_BranchProc)oms_BranchInPlace;
  oms_procsetnum = z_InstallProcSet((z_ProcSetP) &oms_procset);
  
#ifdef USE_PROFILE
  prof_zram_init();
#endif /* USE_PROFILE */

} /* initproc */

#endif


na_node_p oms_init_local(na_node_p local, params_p params){


  return(local);

}

bool oms_read_boundary_hook(na_node_p global, na_node_p local, 
			    int b, int sigma)
{
  be_block_p bdy_eq=na_access(global,SLOT_bdy_eq);
  array_index_t    container=be_bindex(bdy_eq,b,0);
  nu_sign_set_p choice_Q=na_access(local,SLOT_choice_Q);
  
  err_debug(DB_STARTUP,("oms","saving choice %d",container));
  (void)nss_add(choice_Q,container,SIGN_ANY); /* enqueue some basis contining b */
  return TRUE;
}

static na_node_p oms_alloc_local(na_node_p global,na_node_p local){
 
  params_p params=na_access(global,SLOT_params);

  int n=params->n;
  int r=params->r;
  int num_bases=binomial(n,r);
  int num_eqns=binomial(n,r+2)*binomial(r+2,4); 

  local=sa16_nalloc(local,SLOT_pe_state, num_eqns, num_eqns );
  local=nss_nalloc(local,SLOT_choice_Q,num_bases);
  na_fixup(local);

  {
    sa16_array_p   pe_state=na_access(local,SLOT_pe_state);
    int i;

    for(i=0; i<sa16_size(pe_state); i++)
      sa16_init(pe_state,i,0);
  }
  {
    lparams_p lparams=na_access(local,SLOT_lparams);

    lparams->depth=0;
    lparams->subtreesize=0;
  }

  /* What is this about? I guess we want to save memory here,
     but destroy seems inconsistent with returning it.
  if (OPT_var_stack)
    na_destroy(local);
  */

  return local;
}

void SaveState(search_state state){

  lparams_p lparams=na_access(state,SLOT_lparams);
  lparams->depth++;

  rg_save(na_access(state,SLOT_ridge_graph));
  sa16_save(na_access(state,SLOT_pe_state));
  sa_save(na_access(state,SLOT_pos_mask));
  sa_save(na_access(state,SLOT_neg_mask));
  sa_save(na_access(state,SLOT_chirotope));

  sa_save(na_access(state,SLOT_boundary));
}


extern void RestoreState(search_state state){

  lparams_p lparams;

  NODECHECK(state);

  lparams=na_access(state,SLOT_lparams);

  lparams->depth--;

   rg_restore(na_access(state,SLOT_ridge_graph));
   sa16_restore(na_access(state,SLOT_pe_state));
   sa_restore(na_access(state,SLOT_pos_mask));
   sa_restore(na_access(state,SLOT_neg_mask));
   sa_restore(na_access(state,SLOT_chirotope));
   sa_restore(na_access(state,SLOT_boundary));
 }
